

import subprocess
import tempfile
from enum import Enum
import re
import sys
import questionary
import click
from subprocess_tee import run
import click_spinner
from . import chatter
from . import config


def extract_arguments(code):
    matches = re.findall(r"\$[0-9]+", code)
    return matches


def require_args_if_need(code):
    args = extract_arguments(code)
    if len(args) > 0:
        args_string = questionary.text(f"Please input arguments:").ask()
        return split_by_whitespace(args_string)
    return []


def execute_bash_tee(filename, args=[]):
    return run(["bash", filename] + args)


def execute_bash(filename, args=[]):
    return subprocess.run(["bash", filename] + args)


def create_tmp_file(code):
    with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
        tmp_file.write(code.encode("utf-8"))
        return tmp_file.name


def split_by_whitespace(string):
    regex = re.compile(r'\s+')
    return regex.split(string)


class InteractiveAgent():
    """Spinnerなどのインタラクションや再実行などを中継するクラス"""

    def __init__(self, agent):
        self.agent = agent

    # ユーザーの質問に対する回答を取得するメソッド
    def think(self, question):
        click.secho("...WANNA is thinking bash script:",  fg="blue")
        with click_spinner.spinner():
            return self.agent.think(question)

    # 回答が不正解だった場合に再度回答を求めるメソッド
    def rethink(self, result):
        click.secho("...WANNA is thinking about executed result",  fg="blue")
        with click_spinner.spinner():
            return self.agent.rethink(result)

    # 変数名の候補を提案するメソッド
    def propose_names(self):
        click.secho("...WANNA is thinking about script names",  fg="blue")
        with click_spinner.spinner():
            return self.agent.propose_names()

    # コードの概要文を提案するメソッド
    def propose_summary(self):
        click.secho("...WANNA is thinking about script description",  fg="blue")
        with click_spinner.spinner():
            return self.agent.propose_summary()

    # 生成されたコードを取得するメソッド
    def get_code(self):
        return self.agent.code

    # コメントを取得するメソッド
    def get_comment(self):
        return self.agent.display_comment


class NextAction(Enum):
    DO = "Do"
    SAVE = "Save"
    ADDITONAL_REQUEST = "Additonal Request"
    ANOTHER_QUESTION = "Another Question"
    EXIT = "Exit"


def call_whats_next():
    na = questionary.select(
        "What's next?",
        choices=[e.value for e in NextAction],
    ).ask()
    return NextAction(na)


def try_save(agent):
    names = agent.propose_names()

    selected_name = questionary.select(
        "I thought of the following names. Which one do you like?",
        choices=names,
    ).ask()
    summary = agent.propose_summary()
    click.echo(f"Saved as {selected_name}: Description:{summary}")
    config.save_script(selected_name, agent.get_code(), summary)


def conversation_cycle(agent, is_display_comment=True):
    display_comment = agent.get_comment()
    code = agent.get_code()

    WANNA_STYLE = {"bold": True, "fg": "blue", "reverse": False}

    if is_display_comment:
        if code is None:
            click.secho("AI Answer:", bold=True, underline=True, reverse=True)
            click.echo(display_comment)
            click.secho(
                "Sorry, I could not generate the proper code from your comment.", fg="red", bold=True)
            sys.exit(0)

        click.secho("WANNA:", **WANNA_STYLE)
        click.echo(display_comment)

    next_action = call_whats_next()

    if next_action == NextAction.DO:
        """実行する場合：実行結果を元に内省して間違いがあれば新しいコードを提案する"""
        tmp = create_tmp_file(code)
        args = require_args_if_need(code)
        result = execute_bash_tee(tmp, args)
        (has_fix, message) = agent.rethink(result)
        if has_fix:
            click.secho("WANNA Reflection (RETHINK):", **WANNA_STYLE)
            return conversation_cycle(agent, is_display_comment=True)
        else:
            click.secho("WANNA Reflection (SUCCESS):", **WANNA_STYLE)
            click.echo(message)
        return conversation_cycle(agent, is_display_comment=False)

    elif next_action == NextAction.SAVE:
        """保存する場合:script名を提案して、概要文を生成して保存する"""
        return try_save(agent)

    elif next_action == NextAction.ADDITONAL_REQUEST:
        """新しい指示を追加する場合:指示を元にコードを提案する"""
        text = questionary.text("What additional requests do you have?").ask()
        agent.think(text)
        return conversation_cycle(agent)

    elif next_action == NextAction.ANOTHER_QUESTION:
        """新しい指示に切り替える場合：これまでのサマリーをデフォルトに再度指示を仰ぐ"""
        return what_wanna_do(agent.propose_summary())

    elif next_action == NextAction.EXIT:
        """辞める場合:exitする"""
        sys.exit(0)


def what_wanna_do(default_question=""):
    question = questionary.text(
        'What do you want to do ?', default=default_question).ask()

    if len(question.strip()) == 0:
        click.secho("Please tell me what you wanna do", fg="red")
        sys.exit(0)
    agent = InteractiveAgent(chatter.BashAgent())
    agent.think(question)
    conversation_cycle(agent)


@ click.group()
def cmd():
    """Shell command launcher with natural language"""
    pass


def fill_command(command):
    if config.exists(command):
        return command
    ideas = config.ideas()
    command_and_question = questionary.autocomplete(
        'Choose command',
        choices=ideas).ask()
    got_command = command_and_question.split(":")[0].strip()
    if config.exists(got_command):
        return got_command
    else:
        sys.exit(0)


def fill_args(cmd):
    return require_args_if_need(config.get_code(cmd))


@ cmd.command()
@ click.argument('command', required=False)
@ click.argument("args", nargs=-1)
def do(command=None, args=None):
    """Execute a command"""
    cmd = fill_command(command)
    args = [a for a in args] if args is not None else []
    filled_args = args if command else fill_args(cmd)

    execute_bash(config.get_command_path(cmd), filled_args)


@ cmd.command()
@ click.argument('question', required=False)
@ click.option("--model", required=False, default="gpt-3.5-turbo")
def think(model, question=None):
    chatter.gpt_model = model
    """Generate a bash script that answers the incoming request"""
    if (question is None):
        what_wanna_do()
    else:
        what_wanna_do(question)


@ cmd.command()
@ click.option('--command', is_flag=True)
def list(command):
    """List up all commands"""

    target_list = config.commands() if command else config.ideas()

    for value in target_list:
        click.echo(value)


@ cmd.command()
@ click.argument('command', required=False)
@ click.argument("args", nargs=-1)
def remove(command, args):
    """Remove selected command"""
    click.echo("Removed:" + command)
    cmd = fill_command(command)
    config.remove(cmd)


def chat_loop(agent):
    command = questionary.text("You").ask()
    if command == "exit":
        sys.exit(0)
    agent.add_user_message(command)
    with click_spinner.spinner():
        message = agent.chat()
        click.echo("Bashbot: " + message)
    chat_loop(agent)


@ cmd.command()
@ click.option("--model", required=False, default="gpt-3.5-turbo")
def chat(model):
    """Chat with bashbot"""
    chatter.gpt_model = model
    agent = chatter.BaseAgent()
    chat_loop(agent)


def main():
    cmd()


if __name__ == '__main__':
    main()



import os
import openai
import json
import re
import sys
from subprocess_tee import run
from glom import glom
from . import system
from . import codedisplay


if "OPENAI_API_KEY" not in os.environ:
    print("ERROR: OPENAI_API_KEY environment variable not found.")
    sys.exit(1)


openai.api_key = os.environ["OPENAI_API_KEY"]

gpt_model = ""


def extract_json_block(text):
    code_regex = r"```json([\s\S]*?)```"
    code_matches = re.findall(code_regex, text)

    if code_matches:
        return code_matches[0]
    else:
        return text


def parse_json(text):
    return json.loads(extract_json_block(text))


def extract_head_tail(text):
    if len(text) <= 100:
        return text
    else:
        head = text[:50]
        tail = text[-50:]
        return head + "..." + tail


# BaseAgentの作成
class BaseAgent():
    def __init__(self):
        self.messages = []
        self.temperature = 0.01

    def add_user_message(self, message):
        self.messages.append({"role": "user", "content": message})

    def add_system_message(self, message):
        self.messages.append({"role": "system", "content": message})

    def add_assistant_message(self, message):
        self.messages.append({"role": "assistant", "content": message})

    def chat(self):
        response = openai.ChatCompletion.create(
            model=gpt_model,
            messages=self.messages,
            temperature=self.temperature,
        )
        message = glom(response, "choices.0.message.content", default=None)
        self.add_assistant_message(message)
        return message


def remove_non_alpha(input_string):
    return re.sub(r'[^a-zA-Z]', '', input_string)


def guess_language(question):
    agent = BaseAgent()

    agent.add_system_message(
        f"""
        guess the language of following text:
        (e.g ja,en,fr,es,zh)

        <Example>
        input : こんにちは
        output: ja

        input : hello
        output : en
        </Example>

        input: {question}
        output:
        """
    )
    message = agent.chat()
    return remove_non_alpha(message)


class BashAgent(BaseAgent):

    def __init__(self, temperature=0.01):
        super().__init__()
        self.temperature = temperature
        self.language = None
        self.messages = []
        self.display_comment = ""
        self.question = []
        self.names = []
        self.code = ""

    def user_language_prompt(self):
        return f"Output in the following language:{self.language}"

    def bash_script_prompt(self):
        return f"""
        If someone asks you to perform a task,
        your job is to come up with a series of bash scripts that will perform that task.
        Do not use anything but bash scripts.Take up to two arguments if necessary.
        Make the script as compatible as possible.
        Use the following format and try to explain your reasoning step by step:


        <Example>
        Q: "Copy files in a directory named "target" to a new directory at the same level as target named "myNewDirectory"."

        The following actions are required
        - Create a new directory
        - Copy files from the first directory to the second
        ```bash
        #!/usr/bin/env bash
        mkdir myNewDirectory
        cp -r target/* myNewDirectory
        ````
        </Example>

        {self.user_language_prompt}
        """

    def listup_filename_prompt(self):
        return """
        If you were to name this bashscript, what file name would you give it?
        Be sure to output 4 candidates
        Please follow the example name without the ".sh" and answer with a list of four choices in json format.

        ```json
        {
            "filename_candidates": [
                "top_mem_processes",
                "high_mem_procs",
                "mem_usage_top5",
                "processes_by_mem"
            ] // should be 4 candicates
        }
        ```
        """

    def reflection_prompt(self):
        return f"""
        実行結果を見て問題があれば、コードの修正を提案せよ。問題なく動作していれば、問題なしと回答せよ。
        """

    def summary_prompt(self):
        return f"""
        Please write 1 line description of this bash script.
        Condition : Must be less than 150 characters.
        Condition : The description of the operation should be simple.
        Condition : The description should be in the form of a sentence.
        {self.user_language_prompt()}
        """

    def rethink(self, result, retry=0):
        if retry >= 3:
            print("Overlimit retry count")
            sys.exit(1)

        self.add_system_message(f"""
        The execution result of your proposed script is as follows Based on the results:

        # RESULT
        returncode : {result.returncode}
        stdout :
        {extract_head_tail(result.stdout)}
        stderr : 
        {extract_head_tail(result.stderr)}

        If there are any problems with the execution results, suggest modifications to the code. If it works fine, reply that there is no problem.
        {self.user_language_prompt()}
        """)
        message = self.chat()
        code = codedisplay.extract_code_block(message)
        if code is not None:
            self.display_comment = codedisplay.highlight_code_block(message)
            self.code = codedisplay.extract_code_block(message)
            return (True, None)
        else:
            return (False, message)

    def think(self, question):
        if self.language is None:
            self.language = guess_language(question)
            self.add_system_message(
                "The operating environment for bash is as follows:"+system.info())
            self.add_system_message(self.bash_script_prompt())

        self.question.append(question)
        self.add_user_message(question)
        message = self.chat()
        self.display_comment = codedisplay.highlight_code_block(message)
        self.code = codedisplay.extract_code_block(message)
        return message

    def propose_names(self, retry=0):
        if retry >= 3:
            print("Overlimit retry count")
            sys.exit(1)
        self.add_system_message(self.listup_filename_prompt())
        message = self.chat()
        try:
            json = parse_json(message)
            self.names = json["filename_candidates"]
            if len(self.names) != 4:
                print("json length error")
                raise Exception("json error")
        except:
            self.messages = self.messages[:-2]
            print(f"retry{retry}")
            print(f"{message}")
            return self.propose_names(retry + 1)

        return self.names

    def propose_summary(self):
        self.add_system_message(self.summary_prompt())
        message = self.chat()
        return message

    def reset(self):
        self.__init__()



import os
import toml

# ホームディレクトリのパスを取得する
home_dir = os.path.expanduser("~")
# `.wanna`ディレクトリのパスを作成する
wanna_dir = os.path.join(home_dir, ".wanna")

# `.wanna`ディレクトリが存在しない場合は作成する
if not os.path.exists(wanna_dir):
    os.mkdir(wanna_dir)


def read_toml_file(file_path):
    with open(file_path, 'r') as f:
        data = toml.load(f)
    return data


def write_toml_file(file_path, data):
    with open(file_path, 'w') as f:
        toml.dump(data, f)


toml_file_name = os.path.join(wanna_dir, "config.toml")
if not os.path.exists(toml_file_name):
    write_toml_file(toml_file_name, {})

config = read_toml_file(toml_file_name)


def add_to_config(name, code, question):
    config[name] = {"question": question, "code": code}
    write_toml_file(toml_file_name, config)


def remove(name):
    del config[name]
    write_toml_file(toml_file_name, config)
    os.remove(os.path.join(wanna_dir, name))


def save_script(name, code, question):
    # configにもデータを保存する
    add_to_config(name, code, question)

    with open(os.path.join(wanna_dir, name), "w") as f:
        f.write(code)

    os.chmod(os.path.join(wanna_dir, name), 0o755)


def ideas():
    return [
        f"{key:<20}:{config[key]['question']}" for key in sorted(config)
    ]


def commands():
    return [
        key for key in sorted(config)
    ]


def exists(command):
    return command in config


def get_code(command):
    return config[command]["code"]


def get_command_path(command):
    return os.path.join(wanna_dir, command)
